#version 150

// This came from some tutorial, mentioning some drivers required this to function
//   properly but did not delve into why. 
precision highp float;

in vec3 fsNormal;
in vec4 fsColor;
in vec2 fsTextureUV;
///For shadowed particles...
in vec4 shadowCoordinates[4];

out vec4 finalFragColor;

// sAlbedoTexture is a 2D texture sampler for diffuse texture.
uniform sampler2D sAlbedoTexture;
uniform vec4 ambientLightColor = vec4(1.0, 1.0, 1.0, 1.0);
uniform vec4 sunLightColor = vec4(1.0, 1.0, 1.0, 1.0);
uniform vec4 sunLightDirection = vec4(-0.447766751, 0.373415232, 0.812444329, 0.422158539);  //Expected to be a unit vector and in world space already.
uniform vec4 uMaterialColor = vec4(1.0, 1.0, 1.0, 1.0);
//These uniforms control the bias of each cascaded map.
uniform int poissonSamples = 8;
uniform sampler2DShadow shadowTexture;
uniform vec4 poissonScales = vec4(1500.0, 2000.0, 2250.0, 2500.0);
uniform vec4 shadowBias = vec4(0.0, 0.0, 0.0001, 0.0002);

float sample_csm_viewports()
{
	float shadowFactor = 1.0;
	float margin = 0.005; //Can be made accurate to 1 texel

	for (int i = 0; i < 4; ++i)
	{
		vec4 biasedShadowCoordinate = shadowCoordinates[i];
		biasedShadowCoordinate.z -= shadowBias[i];

		vec2 texCoord = biasedShadowCoordinate.xy / biasedShadowCoordinate.w;
		bool outside = any(greaterThan(texCoord.xy, vec2(1.0 - margin))) || any(lessThan(texCoord.xy, vec2(margin)));
		if (false == outside)
		{
			//TODO: Send the crop matrices as uniforms.
			float x = float(i % 2);
			float y = float(i / 2);
			mat4 crop = mat4(
				0.5, 0.0, 0.0, x * 0.5,
				0.0, 0.5, 0.0, y * 0.5,
				0.0, 0.0, 1.0, 0.0,
				0.0, 0.0, 0.0, 1.0
			);

			shadowFactor = 0.0;
			vec4 shadowCoords = crop * biasedShadowCoordinate;
			shadowFactor += textureProj(shadowTexture, shadowCoords);
			break;
		}
	}

	return shadowFactor;
}

void main(void)
{
	float shadowFactor = sample_csm_viewports();

	float nDotL = abs(dot(fsNormal, vec3(sunLightDirection)));

	vec4 textureColor = texture(sAlbedoTexture, fsTextureUV);
	vec4 premultMaterialColor = vec4(uMaterialColor.rgb, 1.0) * (textureColor.a * fsColor.a);
	vec3 finalColor = vec3(fsColor * premultMaterialColor * textureColor) * max(0.225, shadowFactor);

	float gamma = 2.2;
	finalFragColor = vec4(pow(finalColor, vec3(1.0 / gamma, 1.0 / gamma, 1.0 / gamma)), textureColor.a * fsColor.a);

	if (finalFragColor.a < 0.05)
	{
		discard;
	}
}
